const OP = require('./op.js');

const PRECEDENCE = {
  "#": 0, // Precedence 0 should be ignored!
  "=": 1,
  "<":7, "<=":7, ">":7, ">=":7, "==":7, "!=":7,
  "+": 10, "-": 10,
  "*": 20, "/": 20
};

function TokenStream(input){
  var pos = 0;

  function peek(){
    return (pos < input.length) ? input[pos] : null;
  }

  function next(){
    if (pos < input.length){
      let v = input[pos];
      pos += 1;
      return v;
    }
    return null;
  }

  function line(){
    return (pos < input.length) ? input[pos].line : -1;
  }

  function col(){
    return (pos < input.length) ? input[pos].col : -1;
  }

  function eol(){
    return (pos > 0) ? (input[pos-1].line !== input[pos].line) : true;
  }

  function eof(){
    return (pos >= input.length);
  }

  function die(msg){
    throw new Error(msg + " Line: " + input[pos].line + ", Col: " + input[pos].col);
  }

  return {
    peek: peek,
    next: next,
    line: line,
    col: col,
    eol: eol,
    eof, eof,
    die: die
  };
}
var stream = null;

function isTokenType(type, val){
  let t = stream.peek();
  return (t && t.type === type && (!val || t.val === val) && t);
}

function isPunc(ch){
  return isTokenType("punc", ch);
}

function isOpCode(ch){
  return isTokenType("opcode", ch);
}

function isLabel(ch){
  return isTokenType("label", ch);
}

function isDirective(ch){
  return isTokenType("directive", ch);
}

function isOp(ch){
  return isTokenType("op", ch);
}


function skipPunc(ch){
  if (!isPunc(ch))
    stream.die("Unexpected punctuation '" + ch + "'.");
  stream.next();
}


function skipDirective(ch){
  if (!isDirective(ch))
    stream.die("Unexpected directive '" + ch + "'.");
  stream.next();
}


function parseDelimited(s,e,d,parser){
  let toEOL = (s === null || e === null);
  let a = [];
  let first = true;
  if (!toEOL){skipPunc(s);}
  while (!stream.eof() && ((!toEOL && isPunc(e)) || (toEOL && !stream.eol()))){
    if (first){
      first = false;
    } else {skipPunc(d);}
    a.push(parser());
  }
  if (!toEOL){skipPunc(e);}
  return a;
}

function parseByteDirective(){
  let line = stream.line();
  let col = stream.col();
  stream.next();
  return {
    type: "directive",
    op: ".byte",
    args: parseDelimited(null, null, ",", parseExpression),
    line: line,
    col: col
  }
}


function parseElseIfDirective(){
  skipDirective(".elseif");
  let cond = parseExpression();
  skipDirective(".then");
  let then = parseBlock([".else", ".elif", ".endif"]);
  let f = {
    type: "directive",
    op: "if",
    cond: cond,
    then: then
  }
  if (isDirective(".elif")){
    f["else"] = parseElseIfDirective();
  } else if (isDirective(".else")){
    f["else"] = parseBlock([".endif"]);
  } else if(!isDirective(".endif")){
    stream.die("Expected '.endif' Directive.");
  }
  return f;
}


function parseIfDirective(){
  skipDirective(".if");
  let cond = parseExpression();
  skipDirective(".then");
  let then = parseExpression();
  let f = {
    type: "directive",
    op: "if",
    cond: cond,
    then: then
  };
  if (isDirective(".elif")){
    f["else"] = parseElseIfDirective();  
  } else if (isDirective(".else")){
    f["else"] = parseBlock([".endif"]);
  }
  skipDirective(".endif");
  return f;
}


function parseOpCode(){
  let line = stream.line();
  let col = stream.col();
  let val = stream.next();
  let mode = 0; // Guess between absolute and zero page.
  if (isOp("#")){
    stream.next();
    mode = 1; // Immediate
  } else if (isPunc("(")){
    mode = 2; // Indirect
  } 
  return {
    type: "opcode",
    op: val,
    args: parseDelimited(null, null, ",", parseExpression),
    mode: mode,
    line: line,
    col: col
  };
}

function parseAtom(){
  if (isPunc("(")){
    stream.next();
    let exp = parseExpression();
    if (isPunc(")")){
      stream.next();
      return exp;
    }
  } else if (isOpCode()){
    return parseOpCode();
  } else if (isDirective(".if")){
    return parseIfDirective();
  } else if (isDirective(".bytes")){
    return parseByteDirective();
  }

  let tok = stream.next();
  if (tok.type === "number" || tok.type === "string" || tok.type === "label")
    return tok;
  stream.die("Unexpected token {type:" + tok.type + ", val:'" + tok.val + "'}.");
}


function parseExpression(){
  return maybeCall(function(){
    return maybeBinary(parseAtom(), 0);
  });
}

function parseCall(func){
  return {
    type: "call",
    func: func,
    args: parseDelimited("(", ")", ",", parseExpression)
  };
}

function maybeCall(expr){
  expr = expr();
  return (isOp("(")) ? parseCall(expr) : expr;
}

function maybeBinary(left, pres){
  let tok = isOp();
  if (tok){
    let cpres = PRECEDENCE[tok.val];
    if (cpres > pres){
      stream.next();
      return maybeBinary({
        type: (tok.val === "=") ? "assign" : "binary",
        op: tok.val,
        left: left,
        right: maybeBinary(parseAtom(), cpres),
        line: tok.line,
        col: tok.col
      }, pres);
    }
  }
  return left;
}


function parseBlock(bed){
  let exp = [];
  let isBlockEnd = (t) => {
    return (bed && t.type === 'directive' && bed.indexOf(t.val) >= 0);
  }
  while (!stream.eof()){
    if (isBlockEnd(stream.peek()))
      break;
    let e = parseExpression();
    if (e.type === "label"){
      e = {
        type: "assign",
        op: "=",
        left: e,
        right: "*", // To designate "current program counter".
        line: e.line,
        col: e.col
      }
    }
    exp.push(e);
    if (!stream.eol())
      stream.die("Expected End of Line.");
  }
  return exp;
}

function parseProg(ed){
  let line = stream.line();
  let col = stream.col();
  return {
    type: "prog",
    expressions: parseBlock(),
    line: line,
    col: col
  };
}

function parse(tokens){
  stream = TokenStream(tokens);
  let p = parseProg();
  stream = null;
  return p;
}


module.exports = Object.freeze({
  parse: parse
});



